@@ -3,7 +3,7 @@
 # (C) Daniel Kasak: dan@entropy.homelinux.org
 # See COPYRIGHT file for full license
 
-# See 'man Gtk2::Ex::DBI' for full documentation ... or of course continue reading
+# See 'man Gtk2::Ex::Datasheet::DBI' for full documentation ... or of course continue reading
 
 package Gtk2::Ex::Datasheet::DBI;
 
@@ -32,7 +32,7 @@ use constant{
 };
 
 BEGIN {
-	$Gtk2::Ex::DBI::Datasheet::VERSION = '0.6';
+			$Gtk2::Ex::DBI::Datasheet::VERSION = '0.7';
 }
 
 sub new {
@@ -49,7 +49,8 @@ sub new {
 			sql_order_by		=> $$req{sql_order_by},	# The 'order by' clause of the query
 			treeview		=> $$req{treeview},	# The Gtk2::Treeview to connect to
 			fields			=> $$req{fields},	# Field definitions
-			multi_select		=> $$req{multi_select}	# Boolean to enable multi selection mode
+			multi_select		=> $$req{multi_select},	# Boolean to enable multi selection mode
+			on_apply		=> $$req{on_apply}	# Code that runs *after* each *record* is applied
 	};
 	
 	bless $self, $class;
@@ -57,7 +58,7 @@ sub new {
 	$self->setup_treeview;
 	
 	# Remember the primary key column
-	$self->{primary_key_column} = scalar(@{$self->{fieldlist}}) + 1;
+	$self->{primary_key_column} = scalar(@{$self->{fieldlist}}) + 1 + ( $self->{dynamic_models} || 0 );
 	
 	$self->query;
 	
@@ -73,8 +74,33 @@ sub setup_treeview {
 	my $self = shift;
 	
 	# Cache the fieldlist array so we don't have to continually query the DB server for it
-	my $sth = $self->{dbh}->prepare($self->{sql_select} . " from " . $self->{table} . " where 0=1");
-	$sth->execute;
+	my $sth;
+	
+	eval {
+		$sth = $self->{dbh}->prepare($self->{sql_select} . " from " . $self->{table} . " where 0=1")
+			|| die $self->{dbh}->errstr;
+	};
+	
+	if ($@) {
+		Gtk2::Ex::Dialogs::ErrorMsg->new_and_run(
+								title		=> "Error in Query",
+								text		=> "Database server says:\n$@"
+							);
+		return FALSE;
+	}
+	
+	eval {
+		$sth->execute || die $self->{dbh}->errstr;
+	};
+	
+	if ($@) {
+		Gtk2::Ex::Dialogs::ErrorMsg->new_and_run(
+								title		=> "Error in Query",
+								text		=> "Database server says:\n$@"
+							);
+		return FALSE;
+	}
+	
 	$self->{fieldlist} = $sth->{'NAME'};
 	
 	# If there are no field definitions, then use these fields from the database
@@ -106,17 +132,18 @@ sub setup_treeview {
 	# Now set up the model and columns
 	for my $field ( @{$self->{fields}} ) {
 		
-		if ( !$field->{renderer} || $field->{renderer} eq "text" ) {
+		$field->{column} = $column_no - 1; # The field number is 1 off the column number ( status column )
+		
+		if ( !$field->{renderer} || $field->{renderer} eq "text" || $field->{renderer} eq "number" ) {
 			
 			#$renderer = Gtk2::CellRendererText->new;
 			$renderer = MOFO::CellRendererText->new;
+			$renderer->{column} = $column_no;
 			
 			if ( ! $self->{readonly} ) {
-				$renderer->set( editable => 1 );
+				$renderer->set( editable => TRUE );
 			}
 			
-			$renderer->{column} = $column_no;
-			
 			$self->{columns}[$column_no] = Gtk2::TreeViewColumn->new_with_attributes(
 													$field->{name},
 													$renderer,
@@ -130,51 +157,56 @@ sub setup_treeview {
 			# Add a string column to the TreeStore definition ( recreated when we query() )
 			push @{$self->{ts_def}}, "Glib::String";
 			
-		} elsif ( $field->{renderer} eq "number" ) {
-			
-			$renderer = MOFO::CellRendererSpinButton->new;
-			
-			if ( ! $self->{readonly} ) {
-				$renderer->set( mode => "editable" );
-			}
-			
-			$renderer->set(
-					min	=> $field->{min}	|| 0,
-					max	=> $field->{max}	|| 9999,
-					digits	=> $field->{digits}	|| 0,
-					step	=> $field->{step}	|| 1
-				      );
-			
-			$renderer->{column} = $column_no;
-			
-			$self->{columns}[$column_no] = Gtk2::TreeViewColumn->new_with_attributes(
-													$field->{name},
-													$renderer,
-													'value'	=> $column_no
-												);
-			
-			$renderer->signal_connect( edited => sub { $self->process_text_editing( @_ ); } );
-			
-			$self->{treeview}->append_column($self->{columns}[$column_no]);
-			
-			# Add a numeric field to the TreeStore definition ( recreated when we query() )
-			push @{$self->{ts_def}}, "Glib::Double";
+		#} elsif ( $field->{renderer} eq "number" ) {
+		#	
+		#	$renderer = MOFO::CellRendererSpinButton->new;
+		#	
+		#	if ( ! $self->{readonly} ) {
+		#		$renderer->set( mode => "editable" );
+		#	}
+		#	
+		#	$renderer->set(
+		#			min	=> $field->{min}	|| 0,
+		#			max	=> $field->{max}	|| 9999,
+		#			digits	=> $field->{digits}	|| 0,
+		#			step	=> $field->{step}	|| 1
+		#		      );
+		#	
+		#	$renderer->{column} = $column_no;
+		#	
+		#	$self->{columns}[$column_no] = Gtk2::TreeViewColumn->new_with_attributes(
+		#											$field->{name},
+		#											$renderer,
+		#											'value'	=> $column_no
+		#										);
+		#	
+		#	$renderer->signal_connect( edited => sub { $self->process_text_editing( @_ ); } );
+		#	
+		#	$self->{treeview}->append_column($self->{columns}[$column_no]);
+		#	
+		#	# Add a numeric field to the TreeStore definition ( recreated when we query() )
+		#	push @{$self->{ts_def}}, "Glib::Double";
 			
 		} elsif ( $field->{renderer} eq "combo" ) {
 			
 			$renderer = Gtk2::CellRendererCombo->new;
+			$renderer->{column} = $column_no;
 			
 			if ( ! $self->{readonly} ) {
+				
 				$renderer->set(
 						editable	=> TRUE,
-						model		=> $field->{model},
 						text_column	=> 1,
 						has_entry	=> TRUE
-						);
+					      );
+				
+				# It's possible that we won't have a model at this point
+				if ( $field->{model} ) {
+					$renderer->set( model	=> $field->{model} );
+				}
+				
 			}
 			
-			$renderer->{column} = $column_no;
-			
 			$self->{columns}[$column_no] = Gtk2::TreeViewColumn->new_with_attributes(
 													$field->{name},
 													$renderer,
@@ -190,6 +222,53 @@ sub setup_treeview {
 			# Add a string column to the TreeStore definition ( recreated when we query() )
 			push @{$self->{ts_def}}, "Glib::String";
 			
+		} elsif ( $field->{renderer} eq "dynamic_combo" ) {
+			
+			$renderer = Gtk2::CellRendererCombo->new;
+			$renderer->{column} = $column_no;
+			
+			# For a dynamic combo, we have to tell the TreeViewColumn where the model is.
+			# Therefore we need to keep track of how many models we've got.
+			# We can't use $self->column_from_name() because this only works for columns that have a matching
+			# field in our SQL command ( ie are in $self->{fieldlist} ). We also have to be careful not to
+			# upset the order of columns in $self->column_from_name and $self->{fieldlist} ... ie we should
+			# append these models at the end of the the main model, just before the primary key
+			
+			$self->{dynamic_models} ++;
+			$renderer->{dynamic_model_no} = $self->{dynamic_models};
+			$renderer->{dynamic_model_position} = scalar @{$self->{fieldlist}} + 1 + $self->{dynamic_model_no};
+			
+			# Keep this position number in the field has as well
+			$field->{dynamic_model_position} = $renderer->{dynamic_model_position};
+			
+			if ( ! $self->{readonly} ) {
+				$renderer->set(
+						editable	=> TRUE,
+						text_column	=> 1,
+						has_entry	=> TRUE
+					      );
+			}
+			
+			$self->{columns}[$column_no] = Gtk2::TreeViewColumn->new_with_attributes(
+													$field->{name},
+													$renderer,
+													text	=> $column_no,
+													model	=> $renderer->{dynamic_model_position}
+												);
+			
+			$renderer->signal_connect( edited => sub { $self->process_text_editing( @_ ); } );
+			
+			$self->{treeview}->append_column($self->{columns}[$column_no]);
+			
+			$self->{columns}[$column_no]->set_cell_data_func($renderer, sub { $self->render_combo_cell( @_ ); } );
+			
+			# Add a string column to the TreeStore definition ( recreated when we query() )
+			push @{$self->{ts_def}}, "Glib::String";
+			
+			# Add a Gtk2::ListStore column to the TreeStore definition for the model of this combo,
+			# ***BUT*** we can't add it here - queue it until the end of the 'normal' columns ( in the SQL select )
+			push @{$self->{ts_models}}, "Gtk2::ListStore";
+			
 		} elsif ( $field->{renderer} eq "toggle" ) {
 			
 			$renderer = Gtk2::CellRendererToggle->new;
@@ -213,6 +292,28 @@ sub setup_treeview {
 			# Add an integer column to the TreeStore definition ( recreated when we query() )
 			push @{$self->{ts_def}}, "Glib::Boolean";
 			
+		} elsif ( $field->{renderer} eq "date" ) {
+			
+			$renderer = MOFO::CellRendererDate->new;
+			$renderer->{column} = $column_no;
+			
+			if ( ! $self->{readonly} ) {
+				$renderer->set( mode => "editable" );
+			}
+			
+			$self->{columns}[$column_no] = Gtk2::TreeViewColumn->new_with_attributes(
+													$field->{name},
+													$renderer,
+													'date'	=> $column_no
+												);
+			
+			$renderer->signal_connect( edited => sub { $self->process_text_editing( @_ ); } );
+			
+			$self->{treeview}->append_column($self->{columns}[$column_no]);
+			
+			# Add a string column to the TreeStore definition ( recreated when we query() )
+			push @{$self->{ts_def}}, "Glib::String";
+			
 		} elsif ( $field->{renderer} eq "none" ) {
 			
 			print "Adding hidden field " . $field->{name} . "\n";
@@ -239,7 +340,27 @@ sub setup_treeview {
 		
 	}
 	
-	# Add a column for the primary key to the TreeStore definition ... *MUST* have a numberic primary key
+	# Now we've finished the 'normal' columns, we can add any queued dynamic model definitions
+	for my $model_def ( @{$self->{ts_models}} ) {
+		push @{$self->{ts_def}}, $model_def;
+	}
+	
+	# Now that all the columns are set up, loop over them again looking for dynamic models, so we can
+	# set up automatic requerying of models when a column they depend on changes. We *could* have done this
+	# in the above loop, but there's a ( remote ) chance that someone will want to set up a dynamic combo
+	# that depends on a column *after* it ... while I can't see why people would do this, it's easy relatively
+	# easy to accomodate anyway.
+	
+	for my $field ( @{$self->{fields}} ) {
+		if ( $field->{renderer} && $field->{renderer} eq "dynamic_combo" ) {
+			for my $criteria ( @{$field->{model_setup}->{criteria}} ) {
+				push @{($self->{columns}[$self->column_from_name($criteria->{column_name})]->get_cell_renderers)[0]->{dependant_columns}},
+					$field->{column};
+			}
+		}
+	}
+	
+	# Finally, a column for the primary key to the TreeStore definition ... *MUST* have a numberic primary key
 	push @{$self->{ts_def}}, "Glib::Int";
 	
 	# Now set up icons for use in the record status column
@@ -255,6 +376,8 @@ sub setup_treeview {
 		$self->{treeview}->get_selection->set_mode("multiple");
 	}
 	
+	$self->{current_width} = 0; # Prevent warnings
+	
 }
 
 sub render_pixbuf_cell {
@@ -275,25 +398,43 @@ sub render_combo_cell {
 	
 	my $combo_model = $renderer->get("model");
 	
-	# Loop through our combo's model and find a match for the above ID to get our text value
-	my $combo_iter = $combo_model->get_iter_first;
-	my $found_match = FALSE;
-	
-	while ($combo_iter) {
+	if ( $combo_model ) {
 		
-		if ($combo_model->get($combo_iter, 0) == $key_value) {
-			$found_match = TRUE;
-			$renderer->set( text	=> $combo_model->get( $combo_iter, 1 ) );
-			last;
+		# Loop through our combo's model and find a match for the above ID to get our text value
+		my $combo_iter = $combo_model->get_iter_first;
+		my $found_match = FALSE;
+		
+		while ($combo_iter) {
+			
+			# TODO: I'm not really sure if we need to test for the type of $key_value
+			# I can *possibly* remove this...
+			if ( ref($key_value) eq "int" ) {
+				if ($combo_model->get($combo_iter, 0) == $key_value) {
+					$found_match = TRUE;
+					$renderer->set( text	=> $combo_model->get( $combo_iter, 1 ) );
+					last;
+				}
+			} else {
+				if ($combo_model->get($combo_iter, 0) eq $key_value) {
+					$found_match = TRUE;
+					$renderer->set( text	=> $combo_model->get( $combo_iter, 1 ) );
+					last;
+				}
+			}
+			
+			$combo_iter = $combo_model->iter_next($combo_iter);
+			
 		}
 		
-		$combo_iter = $combo_model->iter_next($combo_iter);
+		# If we haven't found a match, default to displaying an empty value
+		if ( !$found_match ) {
+			$renderer->set( text	=> "" );
+		}
+		
+	} else {
+		
+		print "Gtk2::Ex::Datasheet::DBI::render_combo_cell called without a model being attached!\n";
 		
-	}
-	
-	# If we haven't found a match, default to displaying an empty value
-	if ( !$found_match ) {
-		$renderer->set( text	=> "" );
 	}
 	
 	return FALSE;
@@ -312,7 +453,16 @@ sub process_text_editing {
 	# If this is a CellRendererCombo, then we have to look up the ID to match $new_text
 	if ( ref($renderer) eq "Gtk2::CellRendererCombo" ) {
 		
-		my $combo_model = $renderer->get("model");
+		my $combo_model;
+		
+		# If this is a dynamic combo, we can't get the model simply by $render->get("model") because
+		# this is unreliable if the user has clicked outside the current row to end editing.
+		if ( $renderer->{dynamic_model_position} ) {
+			$combo_model = $model->get( $iter, $renderer->{dynamic_model_position} );
+		} else {
+			$combo_model = $renderer->get("model");
+		}
+		
 		my $combo_iter = $combo_model->get_iter_first;
 		my $found_match = FALSE;
 		
@@ -356,6 +506,32 @@ sub process_text_editing {
 		
 		$model->set( $iter, $column_no, $new_text );
 		
+		# If this column has dependant cells ( ie dynamic combos ) pointing at it, refresh them
+		if ( $renderer->{dependant_columns} ) {
+			for my $dependant ( @{$renderer->{dependant_columns}} ) {
+				
+				# Get the current row in an array
+				my @data = $model->get( $model->get_iter ($path) );
+				
+				# We don't want the status column in there - it's not in the SQL fieldlist
+				my $status = shift(@data);
+				
+				# Create a new model
+				my $new_model = $self->create_dynamic_model(
+									$self->{fields}[$dependant]->{model_setup},
+									\@data
+								       );
+				
+				# Dump the combo model in the main TreeView model
+				$model->set(
+						$iter,
+						$self->{fields}[$dependant]->{dynamic_model_position},
+						$new_model
+					   );
+				
+			}
+		}
+		
 	}
 	
 	return FALSE;
@@ -456,19 +632,40 @@ sub query {
 	while (my @row = $sth->fetchrow_array) {
 		
 		my @model_row;
+		my @dynamic_models;
 		my $column = 0;
 		
 		# Append a new treeiter, and the status indicator
 		push @model_row, $liststore->append, STATUS_COLUMN, UNCHANGED;
 		
 		for my $field (@{$self->{fields}}) {
-			push @model_row, $column + 1, $row[$column];
-			$column++;
 			
+			push @model_row, $column + 1, $row[$column];
+			
+			# If this is a dynamic combo, construct it's model now and queue it to be appended
+			# at the end of the 'normal' columns
+			
+			# *** TODO *** Do we need to actually queue this stuff, or can we specify them
+			# out-of-order, as long as we use the right values for the column number?
+			
+			if ( $field->{renderer} && $field->{renderer} eq "dynamic_combo" ) {
+				push @dynamic_models,
+					$field->{dynamic_model_position},
+					$self->create_dynamic_model( $field->{model_setup}, \@row );
+			}
+			
+			$column++;
+		}
+		
+		# Append queued models for dynamic combos
+		for my $dynamic_model ( @dynamic_models ) {
+			push @model_row, $dynamic_model;
 		}
 		
 		# Append the primary key to the end
-		push @model_row, $column + 1, $row[$column];
+		push @model_row,
+			$column + 1 + ( $self->{dynamic_models} || 0 ),
+			$row[$column];
 		
 		$liststore->set(@model_row);
 		
@@ -480,6 +677,14 @@ sub query {
 	
 }
 
+sub undo {
+	
+	my $self = shift;
+	
+	$self->query( undef, TRUE );
+	
+}
+
 sub changed {
 	
 	my ( $self, $liststore, $treepath, $iter ) = @_;
@@ -550,6 +755,7 @@ sub apply {
 			my @values;			# An array of values taken from the current record
 			my $placeholders;		# A string of placeholders, eg ( ?, ?, ? )
 			my $field_index = 1;		# Start at offset=1 to skip over changed flag
+			my $primary_key = undef;	# We pass this to the on_apply() function
 			
 			foreach my $field ( @{$self->{fieldlist}} ) {
 				if ( $status == INSERTED ) {
@@ -558,7 +764,7 @@ sub apply {
 				} else {
 					$sql_fields .= " $field=?,";
 				}
-				push @values, $model->get($iter, $field_index);
+				push @values, $model->get( $iter, $field_index );
 				$field_index++;
 			}
 			
@@ -571,7 +777,8 @@ sub apply {
 			} else {
 				$sql = "update " . $self->{table} . " set $sql_fields"
 					. " where " . $self->{primary_key} . "=?";
-				push @values, $model->get($iter, $field_index);
+				$primary_key = $model->get( $iter, $self->{primary_key_column} );
+				push @values, $primary_key;
 			}
 			
 			my $sth;
@@ -603,7 +810,8 @@ sub apply {
 			
 			# If we just inserted a record, we have to fetch the primary key and replace the current '!' with it
 			if ( $status == INSERTED ) {
-				$model->set($iter, $self->{primary_key_column}, $self->last_insert_id);
+				$primary_key = $self->last_insert_id;
+				$model->set( $iter, $self->{primary_key_column}, $primary_key );
 			}
 			
 			# If we've gotten this far, the update was OK, so we'll reset the 'changed' flag
@@ -612,6 +820,31 @@ sub apply {
 			$model->set($iter, STATUS_COLUMN, UNCHANGED);
 			$model->signal_handler_unblock($self->{changed_signal});
 			
+			# Execute user-defined functions
+			if ( $self->{on_apply} ) {
+				
+				# Better change the status indicator back into text, rather than make
+				# people use our constants. I think, anyway ...
+				my $status_txt;
+				
+				if ( $status == INSERTED ) {
+					$status_txt = "inserted";
+				} elsif ( $status == CHANGED ) {
+					$status_txt = "changed";
+				} elsif ( $status == DELETED ) {
+					$status_txt = "deleted"
+				}
+				
+				# Do people want the whole row? I don't. Maybe others would? Wait for requests...
+				$self->{on_apply}(
+						  {
+							status		=> $status_txt,
+							primary_key	=> $primary_key,
+						  }
+						 );
+				
+			}
+			
 		}
 		
 		$iter = $model->iter_next($iter);
@@ -766,8 +999,115 @@ sub last_insert_id {
 	
 }
 
+sub replace_combo_model {
+	
+	# This function replaces a *normal* combo ( NOT a dynamic one ) with a new one
+	
+	my ( $self, $column_no, $model ) = @_;
+	
+	my $column = $self->{treeview}->get_column($column_no);
+	my $renderer = ($column->get_cell_renderers)[0];
+	$renderer->set( model => $model );
+	
+	return TRUE;
+	
+}
+
+sub create_dynamic_model {
+	
+	# This function accepts a combo definition and a row of data ( *MINUS* the record status column ),
+	# and creates a combo model to insert back into the main TreeView's model
+	# We currently only support a model with 2 columns: an ID column and a Display column
+	
+	# *** TODO *** support adding more columns to the model
+	
+	my ( $self, $model_setup, $data ) = @_;
+	
+	# Firstly we clone the database handle, as the DBD::ODBC / FreeTDS combo won't allow
+	# multiple active statements on the same connection
+	
+	# *** TODO *** Test for the DBD::ODBC driver type so we don't clone the dbh unless we need to
+	
+	my $dbh = $self->{dbh}->clone;
+	
+	my $liststore = Gtk2::ListStore->new(
+						"Glib::String",
+						"Glib::String"
+					    );
+	
+	my $sql = "select " . $model_setup->{id} . ", " . $model_setup->{display} . " from " . $model_setup->{table};
+	my @bind_variables;
+	
+	if ( $model_setup->{criteria} ) {
+		$sql .= " where";
+		for my $criteria ( @{$model_setup->{criteria}} ) {
+			$sql .= " " . $criteria->{field} . "=? and";
+			push @bind_variables, $$data[$self->column_from_name($criteria->{column_name}) - 1];
+		}
+	}
+	
+	$sql = substr( $sql, 0, length($sql) - 3 ); # Remove trailing 'and'
+	
+	if ( $model_setup->{group_by} ) {
+		$sql .= " " . $model_setup->{group_by};
+	}
+	
+	if ( $model_setup->{order_by} ) {
+		$sql .= " " . $model_setup->{order_by};
+	}
+	
+	my $sth;
+	
+	eval {
+		$sth = $dbh->prepare($sql) || die $dbh->errstr;
+	};
+	
+	if ($@) {
+		new_and_run Gtk2::Ex::Dialogs::ErrorMsg(
+					title   => "Error creating combo model!",
+					text    => "DB server says:\n$@"
+				       );
+		return FALSE;
+	}
+	
+	$sth->execute(@bind_variables);
+	
+	my $iter;
+    
+	while (my @record = $sth->fetchrow_array) {
+		
+	        $iter = $liststore->append;
+		$liststore->set(
+					$iter,
+					0, $record[0],
+					1, $record[1]
+			       );
+	        
+	}
+	
+	$sth->finish;
+	$dbh->disconnect;
+	
+	return $liststore;
+	
+}
+
 1;
 
+#######################################################################################
+# That's the end of Gtk2::Ex::Datasheet::DBI
+# What follows is stuff I've plucked from around the place
+#######################################################################################
+
+
+
+
+
+
+#######################################################################################
+# Custom CellRendererText
+#######################################################################################
+
 package MOFO::CellEditableText;
 
 # Copied and pasted from Odot
@@ -928,6 +1268,10 @@ sub START_EDITING {
 	
 }
 
+#######################################################################################
+# CellRendererSpinButton - current non-functional and disabled :(
+#######################################################################################
+
 package MOFO::CellRendererSpinButton;
 
 use POSIX qw(DBL_MAX UINT_MAX);
@@ -1057,8 +1401,8 @@ sub RENDER {
 
 sub START_EDITING {
 	
-	my ($cell, $event, $view, $path, $background_area, $cell_area, $flags) = @_;
-	my $spin_button = Gtk2::SpinButton -> new_with_range($cell -> get(qw(min max step)));
+	my ( $cell, $event, $view, $path, $background_area, $cell_area, $flags ) = @_;
+	my $spin_button = Gtk2::SpinButton -> new_with_range( $cell -> get(qw(min max step)) );
 	
 	$spin_button -> set_value($cell -> get("value"));
 	$spin_button -> set_digits($cell -> get("digits"));
@@ -1138,6 +1482,325 @@ sub START_EDITING {
 1;
 
 
+#######################################################################################
+# CellRendererDate
+#######################################################################################
+
+# Copyright (C) 2003 by Torsten Schoenfeld
+# 
+# This library is free software; you can redistribute it and/or modify it under
+# the terms of the GNU Library General Public License as published by the Free
+# Software Foundation; either version 2.1 of the License, or (at your option)
+# any later version.
+# 
+# This library is distributed in the hope that it will be useful, but WITHOUT
+# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+# FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
+# more details.
+# 
+# You should have received a copy of the GNU Library General Public License
+# along with this library; if not, write to the Free Software Foundation, Inc.,
+# 59 Temple Place - Suite 330, Boston, MA  02111-1307  USA.
+#
+# $Header: /cvsroot/gtk2-perl/gtk2-perl-xs/Gtk2/examples/cellrenderer_date.pl,v 1.5 2005/01/07 21:31:59 kaffeetisch Exp $
+#
+
+
+use strict;
+use Gtk2 -init;
+
+package MOFO::CellRendererDate;
+
+use Glib::Object::Subclass
+  "Gtk2::CellRenderer",
+  signals => {
+    edited => {
+      flags => [qw(run-last)],
+      param_types => [qw(Glib::String Glib::Scalar)],
+    },
+  },
+  properties => [
+    Glib::ParamSpec -> boolean("editable", "Editable", "Can I change that?", 0, [qw(readable writable)]),
+    Glib::ParamSpec -> string("date", "Date", "What's the date again?", "", [qw(readable writable)]),
+  ]
+;
+
+use constant x_padding => 2;
+use constant y_padding => 3;
+
+use constant arrow_width => 15;
+use constant arrow_height => 15;
+
+sub hide_popup {
+  my ($cell) = @_;
+
+  Gtk2 -> grab_remove($cell -> { _popup });
+  $cell -> { _popup } -> hide();
+}
+
+sub get_today {
+  my ($cell) = @_;
+
+  my ($day, $month, $year) = (localtime())[3, 4, 5];
+  $year += 1900;
+  $month += 1;
+
+  return ($year, $month, $day);
+}
+
+sub get_date {
+  my ($cell) = @_;
+
+  my $text = $cell -> get("date");
+  my ($year, $month, $day) = $text
+    ? split(/[\/-]/, $text)
+    : $cell -> get_today();
+
+  return ($year, $month, $day);
+}
+
+sub add_padding {
+  my ($cell, $year, $month, $day) = @_;
+  return ($year, sprintf("%02d", $month), sprintf("%02d", $day));
+}
+
+sub INIT_INSTANCE {
+  my ($cell) = @_;
+
+  my $popup = Gtk2::Window -> new ('popup');
+  my $vbox = Gtk2::VBox -> new(0, 0);
+
+  my $calendar = Gtk2::Calendar -> new();
+
+  my $hbox = Gtk2::HBox -> new(0, 0);
+
+  my $today = Gtk2::Button -> new('Today');
+  my $none = Gtk2::Button -> new('None');
+
+  $cell -> {_arrow} = Gtk2::Arrow -> new("down", "none");
+
+  # We can't just provide the callbacks now because they might need access to
+  # cell-specific variables.  And we can't just connect the signals in
+  # START_EDITING because we'd be connecting many signal handlers to the same
+  # widgets.
+  $today -> signal_connect(clicked => sub {
+    $cell -> { _today_clicked_callback } -> (@_)
+      if (exists($cell -> { _today_clicked_callback }));
+  });
+
+  $none -> signal_connect(clicked => sub {
+    $cell -> { _none_clicked_callback } -> (@_)
+      if (exists($cell -> { _none_clicked_callback }));
+  });
+
+  $calendar -> signal_connect(day_selected_double_click => sub {
+    $cell -> { _day_selected_double_click_callback } -> (@_)
+      if (exists($cell -> { _day_selected_double_click_callback }));
+  });
+
+  $calendar -> signal_connect(month_changed => sub {
+    $cell -> { _month_changed } -> (@_)
+      if (exists($cell -> { _month_changed }));
+  });
+
+  $hbox -> pack_start($today, 1, 1, 0);
+  $hbox -> pack_start($none, 1, 1, 0);
+
+  $vbox -> pack_start($calendar, 1, 1, 0);
+  $vbox -> pack_start($hbox, 0, 0, 0);
+
+  # Find out if the click happended outside of our window.  If so, hide it.
+  # Largely copied from Planner (the former MrProject).
+
+  # Implement via Gtk2::get_event_widget?
+  $popup -> signal_connect(button_press_event => sub {
+    my ($popup, $event) = @_;
+
+    if ($event -> button() == 1) {
+      my ($x, $y) = ($event -> x_root(), $event -> y_root());
+      my ($xoffset, $yoffset) = $popup -> window() -> get_root_origin();
+
+      my $allocation = $popup -> allocation();
+
+      my $x1 = $xoffset + 2 * $allocation -> x();
+      my $y1 = $yoffset + 2 * $allocation -> y();
+      my $x2 = $x1 + $allocation -> width();
+      my $y2 = $y1 + $allocation -> height();
+
+      unless ($x > $x1 && $x < $x2 && $y > $y1 && $y < $y2) {
+        $cell -> hide_popup();
+        return 1;
+      }
+    }
+
+    return 0;
+  });
+
+  $popup -> add($vbox);
+
+  $cell -> { _popup } = $popup;
+  $cell -> { _calendar } = $calendar;
+}
+
+sub START_EDITING {
+  my ($cell, $event, $view, $path, $background_area, $cell_area, $flags) = @_;
+
+  my $popup = $cell -> { _popup };
+  my $calendar = $cell -> { _calendar };
+
+  # Specify the callbacks.  Will be called by the signal handlers set up in
+  # INIT_INSTANCE.
+  $cell -> { _today_clicked_callback } = sub {
+    my ($button) = @_;
+    my ($year, $month, $day) = $cell -> get_today();
+
+    $cell -> signal_emit(edited => $path, join("-", $cell -> add_padding($year, $month, $day)));
+    $cell -> hide_popup();
+  };
+
+  $cell -> { _none_clicked_callback } = sub {
+    my ($button) = @_;
+
+    $cell -> signal_emit(edited => $path, "");
+    $cell -> hide_popup();
+  };
+
+  $cell -> { _day_selected_double_click_callback } = sub {
+    my ($calendar) = @_;
+    my ($year, $month, $day) = $calendar -> get_date();
+
+    $cell -> signal_emit(edited => $path, join("-", $cell -> add_padding($year, ++$month, $day)));
+    $cell -> hide_popup();
+  };
+
+  $cell -> { _month_changed } = sub {
+    my ($calendar) = @_;
+
+    my ($selected_year, $selected_month) = $calendar -> get_date();
+    my ($current_year, $current_month, $current_day) = $cell -> get_today();
+
+    if ($selected_year == $current_year &&
+        ++$selected_month == $current_month) {
+      $calendar -> mark_day($current_day);
+    }
+    else {
+      $calendar -> unmark_day($current_day);
+    }
+  };
+
+  my ($year, $month, $day) = $cell -> get_date();
+
+  $calendar -> select_month($month - 1, $year);
+  $calendar -> select_day($day);
+
+  # Necessary to get the correct allocation of the popup.
+  $popup -> move(-500, -500);
+  $popup -> show_all();
+
+  # Align the top right edge of the popup with the the bottom right edge of the
+  # cell.
+  my ($x_origin, $y_origin) =  $view -> get_bin_window() -> get_origin();
+
+  $popup -> move(
+    $x_origin + $cell_area -> x() + $cell_area -> width() - $popup -> allocation() -> width(),
+    $y_origin + $cell_area -> y() + $cell_area -> height()
+  );
+
+  # Grab the focus and the pointer.
+  Gtk2 -> grab_add($popup);
+  $popup -> grab_focus();
+
+  Gtk2::Gdk -> pointer_grab($popup -> window(),
+                            1,
+                            [qw(button-press-mask
+                                button-release-mask
+                                pointer-motion-mask)],
+                            undef,
+                            undef,
+                            0);
+
+  return;
+}
+
+sub get_date_string {
+  my $cell = shift;
+  return $cell->get ('date');
+}
+
+sub calc_size {
+  my ($cell, $layout) = @_;
+  my ($width, $height) = $layout -> get_pixel_size();
+
+  return (0,
+          0,
+          $width + x_padding * 2 + arrow_width,
+          $height + y_padding * 2);
+}
+
+sub GET_SIZE {
+  my ($cell, $widget, $cell_area) = @_;
+
+  my $layout = $cell -> get_layout($widget);
+  $layout -> set_text($cell -> get_date_string());
+
+  return $cell -> calc_size($layout);
+}
+
+sub get_layout {
+  my ($cell, $widget) = @_;
+
+  return $widget -> create_pango_layout("");
+}
+
+sub RENDER {
+  my ($cell, $window, $widget, $background_area, $cell_area, $expose_area, $flags) = @_;
+  my $state;
+
+  if ($flags & 'selected') {
+    $state = $widget -> has_focus()
+      ? 'selected'
+      : 'active';
+  } else {
+    $state = $widget -> state() eq 'insensitive'
+      ? 'insensitive'
+      : 'normal';
+  }
+
+  my $layout = $cell -> get_layout($widget);
+  $layout -> set_text($cell -> get_date_string());
+
+  my ($x_offset, $y_offset, $width, $height) = $cell -> calc_size($layout);
+
+  $widget -> get_style -> paint_layout($window,
+                                       $state,
+                                       1,
+                                       $cell_area,
+                                       $widget,
+                                       "cellrenderertext",
+                                       $cell_area -> x() + $x_offset + x_padding,
+                                       $cell_area -> y() + $y_offset + y_padding,
+                                       $layout);
+
+  $widget -> get_style -> paint_arrow ($window,
+                                       $widget->state,
+                                       'none',
+                                       $cell_area,
+                                       $cell -> { _arrow },
+                                       "",
+                                       "down",
+                                       1,
+                                       $cell_area -> x + $cell_area -> width - arrow_width,
+                                       $cell_area -> y + $cell_area -> height - arrow_height - 2,
+                                       arrow_width - 3,
+                                       arrow_height);
+}
+
+1;
+
+#######################################################################################
+
+
+
 =head1 NAME
 
 Gtk2::Ex::Datasheet::DBI
@@ -1227,18 +1890,21 @@ Other keys accepted are:
   multi_selcet    - a boolean to turn on the TreeView's 'multiple' selection mode
   fields          - an array of hashes to describe the fields ( columns ) in the TreeView
   
-Each item in the 'fields' key is a hash, with the following keys:
+Each item in the 'fields' key is a hash, with the following possible keys:
   
   name            - the name to display in the column's heading
   x_percent       - a percentage of the available width to use for this column
   x_absolute      - an absolute value to use for the width of this column
   renderer        - string name of renderer - possible values are currently:
-                        - text    - default if no renderer defined )
-			- number  - invokes a customer CellRendererSpin button - see below
-                        - combo   - requires a model to be defined as well )
-                        - toggle  - good for boolean values )
-                        - none    - use this for hidden columns
+                    - text           - default if no renderer defined
+                    - number         - invokes a customer CellRendererSpin button ( diabled and reverts to text )
+                    - combo          - static combo box with a pre-defined list of options
+                    - dynamic_combo  - combo box that depends on values in the current row
+                    - toggle         - good for boolean values
+                    - date           - good for dates - MUST be in YYYY-MM-DD format ( ie most databases should be OK )
+                    - none           - use this for hidden columns
   model           - a TreeModel to use with a combo renderer
+  model_setup     - object describing the setup of a dynamic_combo ( see below )
   validation      - a sub to run after data entry and before the value is accepted to validate data
 
 In the case of a 'number' renderer, the following keys are also used:
@@ -1248,7 +1914,53 @@ In the case of a 'number' renderer, the following keys are also used:
   digits          - the number of decimal places in the spinbutton
   step            - the value that the spinbutton's buttons spin the value by :)
 
-=head2 query ( [ new_where_clause ], [ dont_apply ] )
+Note that as my MOFO::CellRendererSpinButton package is broken ( doesn't accept changes to value ),
+any fields with a number renderer will currently default back to a text renderer. I will re-enable
+the number renderer as soon as I figure out what the problem is. In the meantime, you can always write
+a small function that checks for numeric input and refer to it in your field's 'validation' key.
+
+For dynamic_combo renderers, the 'model_setup' object should take the following form:
+
+{
+
+  id              => "ID"
+  display         => "Description",
+  table           => "SomeTable",
+  criteria        => [
+                        {
+                             field          => "first_where_clause_field",
+                             column_name    => "column_name_of_first_value_to_use"
+                        },
+                        {
+                             field          => "second_where_clause_field",
+                             column_name    => "column_name_of_second_value_to_use"
+                        }
+                     ],
+  group_by        => "group by ID, Description",
+  order_by        => "order by some_field_to_order_by"
+
+}
+
+Briefly ...
+
+The 'id' key defines the primary key in the table you are querying. This is the value that will be
+stored in the dynamic_combo column.
+
+The 'display' key defines the text value that will be *displayed* in the the dynamic_combo column,
+and also in the list of combo options.
+
+The 'table' key is the source table to query.
+
+The 'criteria' key is an array of hashes for you to define criteria. Inside each hash, you have:
+
+  - 'field' key, which is the field in the table you are querying ( ie it will go into the where clause )
+  - 'column_name' key, which is the *SQL* column name to use as limiting value in the where clause
+
+The 'group_by' key is a 'group by' clause. You *shouldn't* need one, but I've added support anyway...
+
+The 'order_by' key is an 'order by' clause
+  
+=head2 query ( [ $new_where_clause ], [ $dont_apply ] )
 
 Requeries the DB server. If there are any outstanding changes that haven't been applied to the database,
 a dialog will be presented to the user asking if they want to apply updates before requerying.
@@ -1260,7 +1972,12 @@ The query method doubles as an 'undo' method if you set the dont_apply flag, eg:
 
 $datasheet->query ( undef, TRUE );
 
-This will requery and reset all the status indicators.
+This will requery and reset all the status indicators. See also undo method, below
+
+=head2 undo
+
+Basically a convenience function that calls $self->query( undef, TRUE ) ... see above.
+I've come to realise that having an undo method makes understanding your code a lot easier later.
 
 =head2 apply
 
@@ -1289,17 +2006,25 @@ Note that you can use the column_from_name method for fetching column numbers fr
 Marks all selected records for deletion, and sets the record status indicator to a 'delete' icon.
 The records will remain in the database until the apply method is called.
 
-=head2 column_from_name ( sql_fieldname )
+=head2 column_from_name ( $sql_fieldname )
 
 Returns a field's column number in the model. Note that you *must* use the SQL fieldname,
 and not the column heading's name in the treeview.
 
-=head2 column_value ( sql_fieldname )
+=head2 column_value ( $sql_fieldname )
 
 Returns the value of the requested column in the currently selected row.
 If multi_select is on and more than 1 row is selected, only the 1st value is returned.
 You *must* use the SQL fieldname, and not the column heading's name in the treeview.
 
+=head2 replace_combo_model ( $column_no, $new_model )
+
+Replaces the model for a combo renderer with a new one.
+You should only use this to replace models for a normal 'combo' renderer.
+An example of when you'd want to do this is if the options in your combo depend on a value
+on your *main* form ( ie not in the datasheet ), and that value changes.
+If you instead want to base your list of options on a value *inside* the datasheet, use
+the 'dynamic_combo' renderer instead ( and don't use replace_combo_model on it ).
 
 =head1 General Ranting
 
@@ -1337,9 +2062,11 @@ Daniel Kasak - dan@entropy.homelinux.org
 
 I think you must be mistaken
 
-=head1 Other cool things you should know about
+=head1 Other cool things you should know about:
 
-This module is part of a 3-some:
+This module is part of an umbrella project, 'Axis Not Evil', which aims to make
+Rapid Application Development of database apps using open-source tools a reality.
+The project includes:
 
 Gtk2::Ex::DBI                 - forms
 
@@ -1347,11 +2074,7 @@ Gtk2::Ex::Datasheet::DBI      - datasheets
 
 PDF::ReportWriter             - reports
 
-Together ( and with a little help from other modules such as Gtk2::GladeXML ),
-these modules give you everything you need for rapid application development of database front-ends
-on Linux, Windows, or ( with a little frigging around ) Mac OS-X.
+All the above modules are available via cpan, or for more information, screenshots, etc, see:
+http://entropy.homelinux.org/axis_not_evil
 
-All the above modules are available via cpan, or from:
-http://entropy.homelinux.org
-
-=head1 Crank ON!
\ No newline at end of file
+=head1 Crank ON!
